/* ----------------------------------------------------------------------
   LAMMPS - Large-scale Atomic/Molecular Massively Parallel Simulator
   www.cs.sandia.gov/~sjplimp/lammps.html
   Steve Plimpton, sjplimp@sandia.gov, Sandia National Laboratories

   Copyright (2003) Sandia Corporation.  Under the terms of Contract
   DE-AC04-94AL85000 with Sandia Corporation, the U.S. Government retains
   certain rights in this software.  This software is distributed under 
   the GNU General Public License.

   See the README file in the top-level LAMMPS directory.
------------------------------------------------------------------------- */

/* ----------------------------------------------------------------------
   Contributing author: Naveen Michaud-Agrawal (Johns Hopkins U)
                        Paul Crozier (SNL)
------------------------------------------------------------------------- */

#include "math.h"
#include "string.h"
#include "fix_spring_rg.h"
#include "atom.h"
#include "update.h"
#include "respa.h"
#include "domain.h"
#include "error.h"

/* ---------------------------------------------------------------------- */

FixSpringRG::FixSpringRG(int narg, char **arg) : Fix(narg, arg)
{
  if (narg != 5) error->all("Illegal fix spring/rg command");

  k = atof(arg[3]);
  rg0_flag = 0;
  if (strcmp(arg[4],"NULL") == 0) rg0_flag = 1;
  else rg0 = atof(arg[4]);
}

/* ---------------------------------------------------------------------- */

int FixSpringRG::setmask()
{
  int mask = 0;
  mask |= POST_FORCE;
  mask |= POST_FORCE_RESPA;
  return mask;
}

/* ---------------------------------------------------------------------- */

void FixSpringRG::init()
{
  // masstotal = total mass of atoms in group

  int *mask = atom->mask;
  int *type = atom->type;
  double *mass = atom->mass;
  int nlocal = atom->nlocal;

  double rmass = 0.0;
  for (int i = 0; i < nlocal; i++)
    if (mask[i] & groupbit) rmass += mass[type[i]];
  MPI_Allreduce(&rmass,&masstotal,1,MPI_DOUBLE,MPI_SUM,world);

  // if rg0 was specified as NULL, compute current Rg
  // only occurs on 1st run

  if (rg0_flag) {
    double com[3];
    rg0 = rg_compute(com);
    rg0_flag = 0;
  }

  if (strcmp(update->integrate_style,"respa") == 0)
    nlevels_respa = ((Respa *) update->integrate)->nlevels;
}

/* ---------------------------------------------------------------------- */

void FixSpringRG::setup()
{
  if (strcmp(update->integrate_style,"verlet") == 0)
    post_force(1);
  else {
    ((Respa *) update->integrate)->copy_flevel_f(nlevels_respa-1);
    post_force_respa(1,nlevels_respa-1,0);
    ((Respa *) update->integrate)->copy_f_flevel(nlevels_respa-1);
  }
}

/* ---------------------------------------------------------------------- */

void FixSpringRG::post_force(int vflag)
{
  // compute current Rg

  double com[3];
  double rg = rg_compute(com);

  // apply restoring force to atoms in group
  // f = -k*(r-r0)*mass/masstotal

  double dx,dy,dz,term1;

  double **f = atom->f;
  double **x = atom->x;
  int *mask = atom->mask;
  int *type = atom->type;
  int *image = atom->image;
  double *mass = atom->mass;
  int nlocal = atom->nlocal;
  
  double massfrac;
  double xprd = domain->xprd;
  double yprd = domain->yprd;
  double zprd = domain->zprd;
   
  int xbox,ybox,zbox;
  for (int i = 0; i < nlocal; i++)
    if (mask[i] & groupbit) {
      term1 = 2.0 * k * (1.0 - rg0/rg);
      xbox = image[i]%1000 - 500;
      ybox = (image[i]/1000) % 1000 - 500;
      zbox = image[i]/1000000 - 500;
      dx = (x[i][0] + xbox*xprd) - com[0];
      dy = (x[i][1] + ybox*yprd) - com[1];
      dz = (x[i][2] + zbox*zprd) - com[2];
      massfrac = mass[type[i]]/masstotal;
      f[i][0] -= term1*dx*massfrac;
      f[i][1] -= term1*dy*massfrac;
      f[i][2] -= term1*dz*massfrac;
    }
}

/* ----------------------------------------------------------------------
   return Rg and com_all[3] = center of mass
------------------------------------------------------------------------- */

double FixSpringRG::rg_compute(double *com_all)
{
  int i;

  // compute center-of-mass
  // use unwrapped coordinates via image
  // else answer is bogus if group straddles a periodic boundary

  double **x = atom->x;
  int *mask = atom->mask;
  int *type = atom->type;
  int *image = atom->image;
  double *mass = atom->mass;
  int nlocal = atom->nlocal;

  double com[3];
  com[0] = com[1] = com[2] = 0.0;
  double xprd = domain->xprd;
  double yprd = domain->yprd;
  double zprd = domain->zprd;
   
  int xbox,ybox,zbox;
  double massone;
  for (i = 0; i < nlocal; i++)
    if (mask[i] & groupbit) {
      xbox = image[i]%1000 - 500;
      ybox = (image[i]/1000) % 1000 - 500;
      zbox = image[i]/1000000 - 500;
      massone = mass[type[i]];
      com[0] += (x[i][0] + xbox*xprd) * massone;
      com[1] += (x[i][1] + ybox*yprd) * massone;
      com[2] += (x[i][2] + zbox*zprd) * massone;
    }
  
  MPI_Allreduce(com,com_all,3,MPI_DOUBLE,MPI_SUM,world);
  com_all[0] /= masstotal;
  com_all[1] /= masstotal;
  com_all[2] /= masstotal;

  // compute radius of gyration

  double rg = 0.0;
  double dx,dy,dz;

  for (i = 0; i < nlocal; i++)
    if (mask[i] & groupbit) {
      xbox = image[i]%1000 - 500;
      ybox = (image[i]/1000) % 1000 - 500;
      zbox = image[i]/1000000 - 500;
      dx = (x[i][0] + xbox*xprd) - com_all[0];
      dy = (x[i][1] + ybox*yprd) - com_all[1];
      dz = (x[i][2] + zbox*zprd) - com_all[2];
      massone = mass[type[i]];
      rg += ((dx*dx + dy*dy + dz*dz) * massone);
    }
  double rg_all;
  MPI_Allreduce(&rg,&rg_all,1,MPI_DOUBLE,MPI_SUM,world);
  
  return sqrt(rg_all/masstotal);
}  

/* ---------------------------------------------------------------------- */

void FixSpringRG::post_force_respa(int vflag, int ilevel, int iloop)
{
  if (ilevel == nlevels_respa-1) post_force(vflag);
}
