/* ----------------------------------------------------------------------
   LAMMPS - Large-scale Atomic/Molecular Massively Parallel Simulator
   www.cs.sandia.gov/~sjplimp/lammps.html
   Steve Plimpton, sjplimp@sandia.gov, Sandia National Laboratories

   Copyright (2003) Sandia Corporation.  Under the terms of Contract
   DE-AC04-94AL85000 with Sandia Corporation, the U.S. Government retains
   certain rights in this software.  This software is distributed under 
   the GNU General Public License.

   See the README file in the top-level LAMMPS directory.
------------------------------------------------------------------------- */

/* ----------------------------------------------------------------------
   Contributing authors: James Fischer (High Performance Technologies, Inc)
                         Vincent Natoli (Stone Ridge Technology)
			 David Richie (Stone Ridge Technology)
------------------------------------------------------------------------- */

#include "math.h"
#include "string.h"
#include "ctype.h"
#include "pair_hybrid.h"
#include "atom.h"
#include "force.h"
#include "pair.h"
#include "neighbor.h"
#include "update.h"
#include "comm.h"
#include "memory.h"
#include "error.h"

#define NEIGHEXTRA 10000

#define MIN(a,b) ((a) < (b) ? (a) : (b))
#define MAX(a,b) ((a) > (b) ? (a) : (b))

/* ---------------------------------------------------------------------- */

PairHybrid::PairHybrid()
{
  nstyles = 0;
}

/* ----------------------------------------------------------------------
   free all arrays 
------------------------------------------------------------------------- */

PairHybrid::~PairHybrid()
{
  if (nstyles) {
    for (int m = 0; m < nstyles; m++) delete styles[m];
    delete [] styles;
    for (int m = 0; m < nstyles; m++) delete [] keywords[m];
    delete [] keywords;
  }

  if (allocated) {
    memory->destroy_2d_int_array(setflag);
    memory->destroy_2d_int_array(map);
    memory->destroy_2d_double_array(cutsq);

    delete [] nnlist;
    delete [] maxneigh;
    for (int m = 0; m < nstyles; m++) memory->sfree(nlist[m]);
    delete [] nlist;

    for (int m = 0; m < nstyles; m++) delete [] firstneigh[m];
    delete [] firstneigh;
    for (int m = 0; m < nstyles; m++) delete [] numneigh[m];
    delete [] numneigh;
  }
}

/* ---------------------------------------------------------------------- */

void PairHybrid::compute(int eflag, int vflag)
{
  int i,j,k,m,n,jfull,nneigh;
  int *neighs,*mapi;
  double **f_orig;

  // save ptrs to original neighbor lists

  int **firstneigh_orig = neighbor->firstneigh;
  int *numneigh_orig = neighbor->numneigh;

  // if this is re-neighbor step, create sub-style lists
  // realloc per-atom per-style firstneigh & numneigh if necessary
  // nnlist[] = length of each sub-style list
  // realloc sub-style nlist if necessary
  // load sub-style list with values from original list

  if (neighbor->ago == 0) {
    int *type = atom->type;
    int nlocal = atom->nlocal;
    int nall = atom->nlocal + atom->nghost;
    if (nlocal > maxlocal) {
      maxlocal = nlocal;
      for (m = 0; m < nstyles; m++) {
	delete [] firstneigh[m];
	delete [] numneigh[m];
      }
      for (m = 0; m < nstyles; m++) {
	firstneigh[m] = new int*[maxlocal];
	numneigh[m] = new int[maxlocal];
      }
    }
    for (m = 0; m < nstyles; m++) nnlist[m] = 0;
    for (i = 0; i < nlocal; i++) {
      mapi = map[type[i]];
      neighs = firstneigh_orig[i];
      nneigh = numneigh_orig[i];
      for (k = 0; k < nneigh; k++) {
	j = neighs[k];
	if (j >= nall) j %= nall;
	nnlist[mapi[type[j]]]++;
      }
    }
    for (m = 0; m < nstyles; m++) {
      if (nnlist[m] > maxneigh[m]) {
	memory->sfree(nlist[m]);
	maxneigh[m] = nnlist[m] + NEIGHEXTRA;
	nlist[m] = (int *)
	  memory->smalloc(maxneigh[m]*sizeof(int),"pair_hybrid:nlist");
      }
      nnlist[m] = 0;
    }
    for (i = 0; i < nlocal; i++) {
      for (m = 0; m < nstyles; m++) {
	firstneigh[m][i] = &nlist[m][nnlist[m]];
	numneigh[m][i] = nnlist[m];
      }
      mapi = map[type[i]];
      neighs = firstneigh_orig[i];
      nneigh = numneigh_orig[i];
      for (k = 0; k < nneigh; k++) {
	j = jfull = neighs[k];
	if (j >= nall) j %= nall;
	m = mapi[type[j]];
	nlist[m][nnlist[m]++] = jfull;
      }
      for (m = 0; m < nstyles; m++)
	numneigh[m][i] = nnlist[m] - numneigh[m][i];
    }
  }
  
  // call each sub-style's compute function
  // for vflag = 1:
  //   sub-style accumulates in its virial[6]
  //   sum sub-style virial[6] to hybrid's virial[6]
  // for vflag = 2:
  //   set atom->f to update->f_pair so sub-style will sum its f to f_pair
  //   call sub-style compute() with vflag % 2 to prevent sub-style
  //     from calling virial_compute()
  //   reset atom->f to stored f_orig
  //   call hybrid virial_compute() which will use update->f_pair
  // must set neighbor->firstneigh/numneigh to sub-style lists before call
  // accumulate sub-style energy,virial in hybrid's energy,virial

  eng_vdwl = eng_coul = 0.0;
  if (vflag) for (n = 0; n < 6; n++) virial[n] = 0.0;

  if (vflag == 2) {
    f_orig = atom->f;
    atom->f = update->f_pair;
  }

  for (m = 0; m < nstyles; m++) {
    neighbor->firstneigh = firstneigh[m];
    neighbor->numneigh = numneigh[m];
    styles[m]->compute(eflag,vflag % 2);
    if (eflag) {
      eng_vdwl += styles[m]->eng_vdwl;
      eng_coul += styles[m]->eng_coul;
    }
    if (vflag == 1) for (n = 0; n < 6; n++) virial[n] += styles[m]->virial[n];
  }

  if (vflag == 2) {
    atom->f = f_orig;
    virial_compute();
  }

  // restore ptrs to original neighbor lists

  neighbor->firstneigh = firstneigh_orig;
  neighbor->numneigh = numneigh_orig;
}

/* ----------------------------------------------------------------------
   allocate all arrays 
------------------------------------------------------------------------- */

void PairHybrid::allocate()
{
  allocated = 1;
  int n = atom->ntypes;

  map = memory->create_2d_int_array(n+1,n+1,"pair:map");

  setflag = memory->create_2d_int_array(n+1,n+1,"pair:setflag");
  for (int i = 1; i <= n; i++)
    for (int j = i; j <= n; j++)
      setflag[i][j] = 0;

  cutsq = memory->create_2d_double_array(n+1,n+1,"pair:cutsq");

  nnlist = new int[nstyles];
  maxneigh = new int[nstyles];
  nlist = new int*[nstyles];
  for (int m = 0; m < nstyles; m++) maxneigh[m] = 0;
  for (int m = 0; m < nstyles; m++) nlist[m] = NULL;

  maxlocal = 0;
  firstneigh = new int**[nstyles];
  numneigh = new int*[nstyles];
  for (int m = 0; m < nstyles; m++) firstneigh[m] = NULL;
  for (int m = 0; m < nstyles; m++) numneigh[m] = NULL;
}

/* ----------------------------------------------------------------------
   create one pair style for each arg in list
------------------------------------------------------------------------- */

void PairHybrid::settings(int narg, char **arg)
{
  int i,m,istyle;

  if (narg < 1) error->all("Illegal pair_style command");

  // delete old lists, since cannot just change settings

  if (nstyles) {
    for (m = 0; m < nstyles; m++) delete styles[m];
    delete [] styles;
    for (m = 0; m < nstyles; m++) delete [] keywords[m];
    delete [] keywords;
  }

  if (allocated) {
    memory->destroy_2d_int_array(setflag);
    memory->destroy_2d_int_array(map);
    memory->destroy_2d_double_array(cutsq);
  }
  allocated = 0;

  // count sub-styles by skipping numeric args
  // one exception is 1st arg of style "table", which is non-numeric word

  nstyles = i = 0;
  while (i < narg) {
    if (strcmp(arg[i],"table") == 0) i++;
    i++;
    while (i < narg && !isalpha(arg[i][0])) i++;
    nstyles++;
  }

  // allocate list of sub-styles

  styles = new Pair*[nstyles];
  keywords = new char*[nstyles];

  // allocate each sub-style and call its settings() with subset of args
  // define subset of sub-styles by skipping numeric args
  // one exception is 1st arg of style "table", which is non-numeric word

  nstyles = i = 0;
  while (i < narg) {
    for (m = 0; m < nstyles; m++)
      if (strcmp(arg[i],keywords[m]) == 0) 
	error->all("Pair style hybrid cannot use same pair style twice");
    if (strcmp(arg[i],"hybrid") == 0) 
      error->all("Pair style hybrid cannot have hybrid as an argument");
    styles[nstyles] = force->new_pair(arg[i]);
    keywords[nstyles] = new char[strlen(arg[i])+1];
    strcpy(keywords[nstyles],arg[i]);
    istyle = i;
    if (strcmp(arg[i],"table") == 0) i++;
    i++;
    while (i < narg && !isalpha(arg[i][0])) i++;
    styles[nstyles]->settings(i-istyle-1,&arg[istyle+1]);
    nstyles++;
  }

  // no more than one "long" sub-style can be used

  int nlong = 0;
  if (force->pair_match("buck/coul/long")) nlong++;
  if (force->pair_match("lj/cut/coul/long")) nlong++;
  if (force->pair_match("lj/cut/coul/long/tip4p")) nlong++;
  if (force->pair_match("lj/charmm/coul/long")) nlong++;
  if (force->pair_match("lj/class2/coul/long")) nlong++;
  if (force->pair_match("dipole/long")) nlong++;
  if (nlong > 1) 
    error->all("Cannot use multiple long-range potentials with pair hybrid");
}

/* ----------------------------------------------------------------------
   set coeffs for one or more type pairs
------------------------------------------------------------------------- */

void PairHybrid::coeff(int narg, char **arg)
{
  if (narg < 3) error->all("Incorrect args for pair coefficients");
  if (!allocated) allocate();

  int ilo,ihi,jlo,jhi;
  force->bounds(arg[0],atom->ntypes,ilo,ihi);
  force->bounds(arg[1],atom->ntypes,jlo,jhi);

  // 3rd arg = pair style name (lj/cut, buck, etc)

  int m;
  for (m = 0; m < nstyles; m++)
    if (strcmp(arg[2],keywords[m]) == 0) break;
  if (m == nstyles) error->all("Pair coeff for hybrid has invalid style");

  // call low-level coeff() for each atom type pair
  // replace 2nd,3rd arg with i,j except some EAM styles use * *
  // call set_coeff with no 1st arg
  // only set the hybrid setflag and map if substyle set the setflag
  //   may not have for some EAM styles

  int count = 0;
  for (int i = ilo; i <= ihi; i++) {
    for (int j = MAX(jlo,i); j <= jhi; j++) {
      if (strcmp(keywords[m],"eam/alloy") == 0 ||
	  strcmp(keywords[m],"eam/sf") == 0) {
	sprintf(arg[1],"%s","*");
	sprintf(arg[2],"%s","*");
      } else {
	sprintf(arg[1],"%d",i);
	sprintf(arg[2],"%d",j);
      }

      styles[m]->coeff(narg-1,&arg[1]);
      if (styles[m]->setflag[i][j]) {
	setflag[i][j] = 1;
	map[i][j] = m;
	count++;
      }
    }
  }

  if (count == 0) error->all("Incorrect args for pair coefficients");
}

/* ----------------------------------------------------------------------
   init for one type pair i,j and corresponding j,i
------------------------------------------------------------------------- */

double PairHybrid::init_one(int i, int j)
{
  // if i,j is set explicity, call its sub-style
  // if i,j is not set and i,i sub-style = j,j sub-style
  //   set map[i][j] to this sub-style and let sub-style do coeff mixing

  double cut;
  if (setflag[i][j]) cut = styles[map[i][j]]->init_one(i,j);
  else if (map[i][i] == map[j][j]) {
    map[i][j] = map[i][i];
    cut = styles[map[i][i]]->init_one(i,j);
  } else error->one("All pair coeffs are not set");

  cutsq[i][j] = cut*cut;
  cutsq[j][i] = cutsq[i][j];
  map[j][i] = map[i][j];
  return cut;
}

/* ----------------------------------------------------------------------
   init specific to this pair style
------------------------------------------------------------------------- */

void PairHybrid::init_style()
{
  for (int m = 0; m < nstyles; m++) styles[m]->init_style();
}

/* ----------------------------------------------------------------------
   proc 0 writes to restart file
------------------------------------------------------------------------- */

void PairHybrid::write_restart(FILE *fp)
{
  fwrite(&nstyles,sizeof(int),1,fp);

  // each sub-style writes its settings

  int n;
  for (int m = 0; m < nstyles; m++) {
    n = strlen(keywords[m]) + 1;
    fwrite(&n,sizeof(int),1,fp);
    fwrite(keywords[m],sizeof(char),n,fp);
    styles[m]->write_restart_settings(fp);
  }
}

/* ----------------------------------------------------------------------
   proc 0 reads from restart file, bcasts
------------------------------------------------------------------------- */

void PairHybrid::read_restart(FILE *fp)
{
  allocate();

  int me = comm->me;
  if (me == 0) fread(&nstyles,sizeof(int),1,fp);
  MPI_Bcast(&nstyles,1,MPI_INT,0,world);

  styles = new Pair*[nstyles];
  keywords = new char*[nstyles];
  
  // each sub-style is created via new_pair() and reads its settings

  int n;
  for (int m = 0; m < nstyles; m++) {
    if (me == 0) fread(&n,sizeof(int),1,fp);
    MPI_Bcast(&n,1,MPI_INT,0,world);
    keywords[m] = new char[n];
    if (me == 0) fread(keywords[m],sizeof(char),n,fp);
    MPI_Bcast(keywords[m],n,MPI_CHAR,0,world);
    styles[m] = force->new_pair(keywords[m]);
    styles[m]->read_restart_settings(fp);
  }
}

/* ---------------------------------------------------------------------- */

void PairHybrid::single(int i, int j, int itype, int jtype,
			double rsq, double factor_coul, double factor_lj,
			int eflag, One &one)
{
  styles[map[itype][jtype]]->
    single(i,j,itype,jtype,rsq,factor_coul,factor_lj,eflag,one);
}

/* ----------------------------------------------------------------------
   modify parameters of the pair style
   simply pass command args to each sub-style of hybrid
------------------------------------------------------------------------- */

void PairHybrid::modify_params(int narg, char **arg)
{
  for (int m = 0; m < nstyles; m++) styles[m]->modify_params(narg,arg);
}

/* ----------------------------------------------------------------------
   memory usage of sub-style firstneigh, numneigh, neighbor list
   add in memory usage of each sub-style itself
------------------------------------------------------------------------- */

int PairHybrid::memory_usage()
{
  int bytes = nstyles*maxlocal * (sizeof(int *) + sizeof(int));
  for (int m = 0; m < nstyles; m++) bytes += maxneigh[m] * sizeof(int);
  for (int m = 0; m < nstyles; m++) bytes += styles[m]->memory_usage();
  return bytes;
}
